import math
import warnings

from contextlib import contextmanager
from multiprocessing.pool import ThreadPool
from queue import Queue
from typing import Callable
from typing import List
from typing import Optional

from sparklightautoml.computations.base import ComputationSlot
from sparklightautoml.computations.base import ComputationsManager
from sparklightautoml.computations.base import ComputationsSession
from sparklightautoml.computations.base import R
from sparklightautoml.computations.base import T
from sparklightautoml.computations.base import logger
from sparklightautoml.computations.utils import (
    duplicate_on_num_slots_with_locations_preferences,
)
from sparklightautoml.computations.utils import get_executors
from sparklightautoml.computations.utils import get_executors_cores
from sparklightautoml.computations.utils import (
    inheritable_thread_target_with_exceptions_catcher,
)
from sparklightautoml.dataset.base import SparkDataset
from sparklightautoml.utils import SparkDataFrame


class ParallelComputationsSession(ComputationsSession):
    def __init__(self, dataset: SparkDataset, parallelism: int, use_location_prefs_mode: int):
        super(ParallelComputationsSession, self).__init__()
        self._parallelism = parallelism
        self._use_location_prefs_mode = use_location_prefs_mode
        self._dataset = dataset
        self._computing_slots: Optional[List[ComputationSlot]] = None
        self._available_computing_slots_queue: Optional[Queue] = None
        self._pool: Optional[ThreadPool] = None
        self._base_pref_locs_df: Optional[SparkDataFrame] = None

    def __enter__(self):
        self._pool = ThreadPool(processes=self._parallelism)
        self._computing_slots = self._make_computing_slots(self._dataset)
        self._available_computing_slots_queue = Queue(maxsize=len(self._computing_slots))
        for cslot in self._computing_slots:
            self._available_computing_slots_queue.put(cslot)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._pool.terminate()
        self._pool = None
        self._available_computing_slots_queue = None
        if self._computing_slots is not None:
            for cslot in self._computing_slots:
                if cslot.dataset is not None:
                    cslot.dataset.unpersist()
        if self._base_pref_locs_df is not None:
            self._base_pref_locs_df.unpersist()

    @contextmanager
    def allocate(self) -> ComputationSlot:
        slot = None
        try:
            assert self._available_computing_slots_queue is not None, "Cannot allocate slots without session"
            slot = self._available_computing_slots_queue.get()
            yield slot
        finally:
            if slot is not None:
                self._available_computing_slots_queue.put(slot)

    def map_and_compute(self, func: Callable[[R], T], tasks: List[R]) -> List[T]:
        assert self._pool is not None
        # TODO: PARALLEL - probably, is not fully correct and needs to be integrated on the thread pool level,
        #  inlcuding one-shot threads
        return self._pool.map(
            lambda task: inheritable_thread_target_with_exceptions_catcher(lambda: func(task))(), tasks
        )

    def compute(self, tasks: List[Callable[[], T]]) -> List[T]:
        assert self._pool is not None
        # TODO: PARALLEL - probably, is not fully correct and needs to be integrated on the thread pool level,
        #  inlcuding one-shot threads
        return self._pool.map(lambda f: inheritable_thread_target_with_exceptions_catcher(f)(), tasks)

    def _make_computing_slots(self, dataset: Optional[SparkDataset]) -> List[ComputationSlot]:
        if dataset is not None and self._use_location_prefs_mode:
            computing_slots = self._make_slots_on_dataset_copies_coalesced_into_preffered_locations(dataset)
        elif dataset is not None:
            computing_slots = self._make_slots_with_current_dataset(dataset)
        else:
            computing_slots = [ComputationSlot(f"{i}") for i in range(self._parallelism)]

        return computing_slots

    def _make_slots_with_current_dataset(self, dataset: SparkDataset) -> List[ComputationSlot]:
        execs = get_executors()
        exec_cores = get_executors_cores()

        num_tasks = max(1, math.floor(len(execs) * exec_cores / self._parallelism))
        num_threads_per_executor = max(1, round(num_tasks / len(execs)))

        computing_slots = [
            ComputationSlot(f"{i}", dataset, num_tasks=num_tasks, num_threads_per_executor=num_threads_per_executor)
            for i in range(self._parallelism)
        ]

        return computing_slots

    def _make_slots_on_dataset_copies_coalesced_into_preffered_locations(
        self, dataset: SparkDataset
    ) -> List[ComputationSlot]:
        logger.warning(
            "Be aware for correct functioning slot-based computations "
            "there should noy be any parallel computations from "
            "different entities (other MLPipes, MLAlgo, etc)."
        )

        execs = get_executors()
        exec_cores = get_executors_cores()

        if math.floor(len(execs) / self._parallelism) > 0 and len(execs) % self._parallelism != 0:
            warnings.warn(
                f"Uneven number of executors per job (one job uses multiple executors): "
                f"{len(execs)} / {self._parallelism}. "
                f"Number of allocated slots may be reduced."
            )

        if math.floor(len(execs) / self._parallelism) == 0 and exec_cores % self._parallelism != 0:
            warnings.warn(
                f"Uneven number of jobs per executor (one job uses only a part of a single executor): "
                f"{exec_cores} / {self._parallelism}. "
                f"Number of allocated slots may be reduced."
            )

        logger.info(
            f"Coalescing dataset into multiple copies (num copies: {self._parallelism}) "
            f"with specified preffered locations"
        )

        if self._parallelism > 1:
            dfs, self._base_pref_locs_df = duplicate_on_num_slots_with_locations_preferences(
                df=dataset.data, num_slots=self._parallelism, enforce_division_without_reminder=False
            )
        else:
            dfs, self._base_pref_locs_df = [dataset.data], None

        assert len(dfs) > 0, "Not dataframe slots are prepared, cannot continue"

        if len(dfs) != self._parallelism:
            logger.warning(
                f"Impossible to allocate desired number of slots, "
                f"probably due to lack of resources: {len(dfs)} < {self._parallelism}"
            )

        # checking all dataframes have the same number of partitions
        unique_num_partitions = set(df.rdd.getNumPartitions() for df in dfs)
        assert len(unique_num_partitions) == 1 and unique_num_partitions.pop() > 0

        num_tasks = dfs[0].rdd.getNumPartitions()
        num_threads_per_executor = min(num_tasks, exec_cores)

        logger.info(f"Continuing with {len(dfs)} prepared slots.")

        dataset_slots = []
        for i, coalesced_df in enumerate(dfs):
            coalesced_dataset = dataset.empty()
            coalesced_dataset.set_data(
                coalesced_df, dataset.features, dataset.roles, name=f"CoalescedForPrefLocs_{dataset.name}"
            )

            # TODO: PARALLEL - add preffered locations logging
            dataset_slots.append(
                ComputationSlot(
                    id=f"{i}",
                    dataset=coalesced_dataset,
                    num_tasks=num_tasks,
                    num_threads_per_executor=num_threads_per_executor,
                )
            )

        return dataset_slots


class ParallelComputationsManager(ComputationsManager):
    def __init__(self, parallelism: int = 1, use_location_prefs_mode: bool = False):
        assert parallelism >= 1
        self._parallelism = parallelism
        self._use_location_prefs_mode = use_location_prefs_mode

    @property
    def parallelism(self) -> int:
        return self._parallelism

    def session(self, dataset: Optional[SparkDataset] = None) -> ParallelComputationsSession:
        return ParallelComputationsSession(dataset, self._parallelism, self._use_location_prefs_mode)



"""Base classes for MLPipeline."""
import uuid
import warnings

from copy import copy
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union
from typing import cast

from lightautoml.dataset.base import RolesDict
from lightautoml.ml_algo.tuning.base import ParamsTuner
from lightautoml.ml_algo.utils import tune_and_fit_predict
from lightautoml.pipelines.ml.base import MLPipeline as LAMAMLPipeline
from lightautoml.pipelines.selection.base import EmptySelector
from pyspark.ml import PipelineModel
from pyspark.ml import Transformer

from ...computations.base import ComputationsSettings
from ...computations.builder import build_computations_manager
from ...dataset.base import SparkDataset
from ...ml_algo.base import SparkTabularMLAlgo
from ...validation.base import SparkBaseTrainValidIterator
from ..base import TransformerInputOutputRoles
from ..features.base import SparkEmptyFeaturePipeline
from ..features.base import SparkFeaturesPipeline
from ..selection.base import SparkSelectionPipelineWrapper


class SparkMLPipeline(LAMAMLPipeline, TransformerInputOutputRoles):
    """Spark version of :class:`~lightautoml.pipelines.ml.base.MLPipeline`. Single ML pipeline.

    Merge together stage of building ML model
    (every step, excluding model training, is optional):

        - Pre selection: select features from input data.
          Performed by
          :class:`~lightautoml.pipelines.selection.base.SelectionPipeline`.
        - Features generation: build new features from selected.
          Performed by
          :class:`~sparklightautoml.pipelines.features.base.SparkFeaturesPipeline`.
        - Post selection: One more selection step - from created features.
          Performed by
          :class:`~lightautoml.pipelines.selection.base.SelectionPipeline`.
        - Hyperparams optimization for one or multiple ML models.
          Performed by
          :class:`~lightautoml.ml_algo.tuning.base.ParamsTuner`.
        - Train one or multiple ML models:
          Performed by :class:`~sparklightautoml.ml_algo.base.SparkTabularMLAlgo`.
          This step is the only required for at least 1 model.

    """

    def __init__(
        self,
        ml_algos: Sequence[Union[SparkTabularMLAlgo, Tuple[SparkTabularMLAlgo, ParamsTuner]]],
        force_calc: Union[bool, Sequence[bool]] = True,
        pre_selection: Optional[SparkSelectionPipelineWrapper] = None,
        features_pipeline: Optional[SparkFeaturesPipeline] = None,
        post_selection: Optional[SparkSelectionPipelineWrapper] = None,
        name: Optional[str] = None,
        persist_before_ml_algo: bool = False,
        computations_settings: Optional[ComputationsSettings] = None,
    ):
        if features_pipeline is None:
            features_pipeline = SparkEmptyFeaturePipeline()

        if pre_selection is None:
            pre_selection = SparkSelectionPipelineWrapper(EmptySelector())

        if post_selection is None:
            post_selection = SparkSelectionPipelineWrapper(EmptySelector())

        super().__init__(ml_algos, force_calc, pre_selection, features_pipeline, post_selection)

        self._output_features = None
        self._output_roles = None
        self._transformer: Optional[Transformer] = None
        self._name = name if name else str(uuid.uuid4())[:5]
        self.ml_algos: List[SparkTabularMLAlgo] = []
        self.pre_selection = cast(SparkSelectionPipelineWrapper, self.pre_selection)
        self.post_selection = cast(SparkSelectionPipelineWrapper, self.post_selection)
        self.features_pipeline = cast(SparkFeaturesPipeline, self.features_pipeline)
        self._milestone_name = f"MLPipe_{self._name}"
        self._input_roles: Optional[RolesDict] = None
        self._output_roles: Optional[RolesDict] = None
        self._persist_before_ml_algo = persist_before_ml_algo
        self._service_columns: Optional[List[str]] = None
        self._computations_manager = build_computations_manager(computations_settings)

    @property
    def input_roles(self) -> Optional[RolesDict]:
        return self._input_roles

    @property
    def output_roles(self) -> Optional[RolesDict]:
        return self._output_roles

    @property
    def name(self) -> str:
        return self._name

    def _build_transformer(self, *args, **kwargs) -> Optional[Transformer]:
        assert self._transformer is not None, f"{type(self)} seems to be not fitted"
        return self._transformer

    def fit_predict(self, train_valid: SparkBaseTrainValidIterator) -> SparkDataset:
        """Fit on train/valid iterator and transform on validation part.

        Args:
            train_valid: Dataset iterator.

        Returns:
            Dataset with predictions of all models.

        """

        # train and apply pre selection
        train_valid = train_valid.apply_selector(self.pre_selection)

        # apply features pipeline
        train_valid = train_valid.apply_feature_pipeline(self.features_pipeline)

        # train and apply post selection
        train_valid = train_valid.apply_selector(self.post_selection)

        with train_valid.frozen() as frozen_train_valid:

            def build_fit_func(ml_algo: SparkTabularMLAlgo, param_tuner: ParamsTuner, force_calc: bool):
                def func():
                    fitted_ml_algo, curr_preds = tune_and_fit_predict(
                        ml_algo, param_tuner, frozen_train_valid, force_calc
                    )
                    fitted_ml_algo = cast(SparkTabularMLAlgo, fitted_ml_algo)
                    curr_preds = cast(SparkDataset, curr_preds)

                    if ml_algo is None:
                        warnings.warn(
                            "Current ml_algo has not been trained by some reason. " "Check logs for more details.",
                            RuntimeWarning,
                        )

                    return fitted_ml_algo, curr_preds

                return func

            fit_tasks = [
                build_fit_func(ml_algo, param_tuner, force_calc)
                for ml_algo, param_tuner, force_calc in zip(self._ml_algos, self.params_tuners, self.force_calc)
            ]

            results = self._computations_manager.compute(fit_tasks)
            ml_algos, preds = [list(el) for el in zip(*results)]
            self.ml_algos.extend(ml_algos)

            assert (
                len(self.ml_algos) > 0
            ), "Pipeline finished with 0 models for some reason.\nProbably one or more models failed"

        del self._ml_algos

        val_preds_ds = SparkDataset.concatenate(
            preds, name=f"{type(self)}_folds_predictions", extra_dependencies=[train_valid]
        )

        self._transformer = PipelineModel(
            stages=[
                # self.pre_selection.transformer(),
                self.features_pipeline.transformer(),
                # self.post_selection.transformer(),
                *[ml_algo.transformer() for ml_algo in self.ml_algos],
            ]
        )

        self._input_roles = copy(train_valid.train.roles)
        self._output_roles = copy(val_preds_ds.roles)
        self._service_columns = train_valid.train.service_columns

        return val_preds_ds

    def predict(self, dataset: SparkDataset) -> SparkDataset:
        """Predict on new dataset.

        Args:
            dataset: Dataset used for prediction.

        Returns:
            Dataset with predictions of all trained models.

        """
        return self._make_transformed_dataset(dataset)

    def _get_service_columns(self) -> List[str]:
        return self._service_columns



import functools
import logging

from abc import ABC
from copy import copy
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union
from typing import cast

from lightautoml.dataset.base import RolesDict
from lightautoml.dataset.roles import NumericRole
from lightautoml.ml_algo.base import MLAlgo
from lightautoml.utils.timer import TaskTimer
from pyspark.ml import Model
from pyspark.ml import PipelineModel
from pyspark.ml import Transformer
from pyspark.ml.param import Params
from pyspark.ml.param.shared import HasInputCols
from pyspark.ml.param.shared import HasOutputCol
from pyspark.ml.param.shared import Param
from pyspark.ml.util import DefaultParamsReadable
from pyspark.ml.util import DefaultParamsWritable
from pyspark.sql import functions as sf

from sparklightautoml.computations.base import ComputationSlot
from sparklightautoml.computations.base import ComputationsManager
from sparklightautoml.computations.base import ComputationsSettings
from sparklightautoml.computations.builder import build_computations_manager
from sparklightautoml.dataset.base import PersistenceLevel
from sparklightautoml.dataset.base import SparkDataset
from sparklightautoml.dataset.roles import NumericVectorOrArrayRole
from sparklightautoml.pipelines.base import TransformerInputOutputRoles
from sparklightautoml.spark_functions import scalar_averaging
from sparklightautoml.spark_functions import vector_averaging
from sparklightautoml.utils import SparkDataFrame
from sparklightautoml.utils import log_exception
from sparklightautoml.validation.base import SparkBaseTrainValidIterator


logger = logging.getLogger(__name__)

SparkMLModel = PipelineModel

ComputationalParameters = Union[Dict[str, Any], ComputationsManager]


class SparkTabularMLAlgo(MLAlgo, TransformerInputOutputRoles, ABC):
    """Machine learning algorithms that accepts numpy arrays as input."""

    _name: str = "SparkTabularMLAlgo"
    _default_validation_col_name: str = SparkBaseTrainValidIterator.TRAIN_VAL_COLUMN

    def __init__(
        self,
        default_params: Optional[dict] = None,
        freeze_defaults: bool = True,
        timer: Optional[TaskTimer] = None,
        optimization_search_space: Optional[dict] = None,
        persist_output_dataset: bool = True,
        computations_settings: Optional[ComputationsSettings] = None,
    ):
        optimization_search_space = optimization_search_space if optimization_search_space else dict()
        super().__init__(default_params, freeze_defaults, timer, optimization_search_space)
        self.n_classes: Optional[int] = None
        self.persist_output_dataset = persist_output_dataset
        # names of columns that should contain predictions of individual models
        self._models_prediction_columns: Optional[List[str]] = None

        self._prediction_role: Optional[Union[NumericRole, NumericVectorOrArrayRole]] = None
        self._input_roles: Optional[RolesDict] = None
        self._service_columns: Optional[List[str]] = None
        self._computations_manager: Optional[ComputationsManager] = build_computations_manager(computations_settings)

    @property
    def features(self) -> Optional[List[str]]:
        """Get list of features."""
        return list(self._input_roles.keys()) if self._input_roles else None

    @features.setter
    def features(self, val: Sequence[str]):
        """List of features."""
        raise NotImplementedError("Unsupported operation")

    @property
    def input_roles(self) -> Optional[RolesDict]:
        return self._input_roles

    @property
    def output_roles(self) -> Optional[RolesDict]:
        return {self.prediction_feature: self.prediction_role}

    @property
    def prediction_feature(self) -> str:
        # return self._prediction_col
        return f"{self._name}_prediction"

    @property
    def prediction_role(self) -> Union[NumericRole, NumericVectorOrArrayRole]:
        return self._prediction_role

    @property
    def validation_column(self) -> str:
        return self._default_validation_col_name

    @property
    def computations_manager(self) -> ComputationsManager:
        return self._computations_manager

    @computations_manager.setter
    def computations_manager(self, value: ComputationsManager):
        self._computations_manager = value

    @log_exception(logger=logger)
    def fit_predict(self, train_valid_iterator: SparkBaseTrainValidIterator) -> SparkDataset:
        """Fit and then predict accordig the strategy that uses train_valid_iterator.

        If item uses more then one time it will
        predict mean value of predictions.
        If the element is not used in training then
        the prediction will be ``numpy.nan`` for this item

        Args:
            train_valid_iterator: Classic cv-iterator.

        Returns:
            Dataset with predicted values.

        """
        logger.info(f"Input columns for MLALgo: {sorted(train_valid_iterator.train.features)}")
        logger.info(f"Train size for MLAlgo: {train_valid_iterator.train.data.count()}")

        assert not self.is_fitted, "Algo is already fitted"

        self._input_roles = copy(train_valid_iterator.train.roles)
        # init params on input if no params was set before
        if self._params is None:
            self.params = self.init_params_on_input(train_valid_iterator)

        iterator_len = len(train_valid_iterator)
        if iterator_len > 1:
            logger.info("Start fitting \x1b[1m{}\x1b[0m ...".format(self._name))
            logger.debug(f"Training params: {self.params}")

        # get metric and loss if None
        self.task = train_valid_iterator.train.task

        valid_ds = cast(SparkDataset, train_valid_iterator.get_validation_data())

        self._infer_and_set_prediction_role(valid_ds)

        self._models_prediction_columns = []

        with train_valid_iterator.frozen() as frozen_train_valid_iterator:
            self.models, preds_dfs, self._models_prediction_columns = self._parallel_fit(
                train_valid_iterator=frozen_train_valid_iterator
            )

        full_preds_df = self._combine_val_preds(train_valid_iterator.get_validation_data(), preds_dfs)
        full_preds_df = self._build_averaging_transformer().transform(full_preds_df)
        full_preds_df = self._build_vector_size_hint(self.prediction_feature, self._prediction_role).transform(
            full_preds_df
        )

        pred_ds = valid_ds.empty()
        pred_ds.set_data(
            full_preds_df,
            list(self.output_roles.keys()),
            self.output_roles,
            dependencies=[train_valid_iterator],
            name=f"{self._name}",
        )
        if self.persist_output_dataset:
            pred_ds = pred_ds.persist(level=PersistenceLevel.REGULAR)

        self._service_columns = train_valid_iterator.train.service_columns

        if iterator_len > 1:
            single_pred_ds = self._make_single_prediction_dataset(pred_ds)
            logger.info(
                f"Fitting \x1b[1m{self._name}\x1b[0m finished. score = \x1b[1m{self.score(single_pred_ds)}\x1b[0m"
            )

        if iterator_len > 1 or "Tuned" not in self._name:
            logger.info("\x1b[1m{}\x1b[0m fitting and predicting completed".format(self._name))

        return pred_ds

    def fit_predict_single_fold(
        self,
        fold_prediction_column: str,
        validation_column: str,
        train: SparkDataset,
        runtime_settings: Optional[Dict[str, Any]] = None,
    ) -> Tuple[SparkMLModel, SparkDataFrame, str]:
        """Train on train dataset and predict on holdout dataset.

        Args:
            fold_prediction_column: column name for predictions made for this fold
            validation_column: name of the column that signals if this row is from train or val
            train: dataset containing both train and val rows.
            runtime_settings: settings important for parallelism and performance that can depend on running processes
            at the moment

        Returns:
            Target predictions for valid dataset.

        """
        raise NotImplementedError

    def predict_single_fold(self, model: SparkMLModel, dataset: SparkDataset) -> SparkDataFrame:
        raise NotImplementedError("Not supported for Spark. Use transformer property instead ")

    def predict(self, dataset: SparkDataset) -> SparkDataset:
        return self._make_transformed_dataset(dataset)

    def _get_service_columns(self) -> List[str]:
        return self._service_columns

    def _infer_and_set_prediction_role(self, valid_ds: SparkDataset):
        outp_dim = 1
        if self.task.name == "multiclass":
            outp_dim = valid_ds.data.select(sf.max(valid_ds.target_column).alias("max")).first()
            outp_dim = outp_dim["max"] + 1
            self._prediction_role = NumericVectorOrArrayRole(
                outp_dim, f"{self.prediction_feature}" + "_{}", force_input=True, prob=True
            )
        elif self.task.name == "binary":
            outp_dim = 2
            self._prediction_role = NumericVectorOrArrayRole(
                outp_dim, f"{self.prediction_feature}" + "_{}", force_input=True, prob=True
            )
        else:
            self._prediction_role = NumericRole(force_input=True)

        self.n_classes = outp_dim

    @staticmethod
    def _get_predict_column(model: SparkMLModel) -> str:
        try:
            return model.getPredictionCol()
        except AttributeError:
            if isinstance(model, PipelineModel):
                return model.stages[-1].getPredictionCol()

            raise TypeError("Unknown model type! Unable ro retrieve prediction column")

    def _predict_feature_name(self):
        return f"{self._name}_prediction"

    def _build_averaging_transformer(self) -> Transformer:
        raise NotImplementedError()

    def _make_single_prediction_dataset(self, dataset: SparkDataset) -> SparkDataset:
        preds = dataset.data.select(SparkDataset.ID_COLUMN, dataset.target_column, self.prediction_feature)
        roles = {self.prediction_feature: dataset.roles[self.prediction_feature]}

        output: SparkDataset = dataset.empty()
        output.set_data(preds, list(roles.keys()), roles, name="single_prediction_dataset")

        return output

    def _combine_val_preds(self, val_data: SparkDataset, val_preds: Sequence[SparkDataFrame]) -> SparkDataFrame:
        # depending on train_valid logic there may be several ways of treating predictions results:
        # 1. for folds iterators - join the results, it will yield the full train dataset
        # 2. for holdout iterators - create None predictions in train_part and join with valid part
        # 3. for custom iterators which may put the same records in
        #   different folds: union + groupby + (optionally) union with None-fied train_part
        # 4. for dummy - do nothing
        assert len(val_preds) > 0

        if len(val_preds) == 1:
            return val_preds[0]

        # we leave only service columns, e.g. id, fold, target columns
        initial_df = cast(SparkDataFrame, val_data[:, []].data)

        full_val_preds = functools.reduce(
            lambda acc, x: acc.join(x, on=SparkDataset.ID_COLUMN, how="left"),
            [val_pred.drop(val_data.target_column) for val_pred in val_preds],
            initial_df,
        )

        return full_val_preds

    def _parallel_fit(
        self, train_valid_iterator: SparkBaseTrainValidIterator
    ) -> Tuple[List[Model], List[SparkDataFrame], List[str]]:
        num_folds = len(train_valid_iterator)

        def _fit_and_val_on_fold(
            fold_id: int, slot: ComputationSlot
        ) -> Optional[Tuple[int, Model, SparkDataFrame, str]]:
            mdl_pred_col = f"{self.prediction_feature}_{fold_id}"
            if num_folds > 1:
                logger.info2(
                    "===== Start working with \x1b[1mfold {}\x1b[0m for \x1b[1m{}\x1b[0m "
                    "=====".format(fold_id, self._name)
                )

            if self.timer.time_limit_exceeded():
                logger.info(f"No time to calculate fold {fold_id}/{num_folds} (Time limit is already exceeded)")
                return None

            runtime_settings = {"num_tasks": slot.num_tasks, "num_threads": slot.num_threads_per_executor}
            train_ds = slot.dataset

            mdl, vpred, _ = self.fit_predict_single_fold(
                mdl_pred_col, self.validation_column, train_ds, runtime_settings
            )
            vpred = vpred.select(SparkDataset.ID_COLUMN, train_ds.target_column, mdl_pred_col)

            return fold_id, mdl, vpred, mdl_pred_col

        results = self.computations_manager.compute_folds(train_valid_iterator, _fit_and_val_on_fold)

        self.timer.write_run_info()

        computed_results = (r for r in results if r is not None)
        _, models, val_preds, model_prediction_cols = [list(el) for el in zip(*computed_results)]

        return models, val_preds, model_prediction_cols


class AveragingTransformer(Transformer, HasInputCols, HasOutputCol, DefaultParamsWritable, DefaultParamsReadable):
    """
    Transformer that gets one or more columns and produce column with average values.
    """

    taskName = Param(Params._dummy(), "taskName", "task name")
    removeCols = Param(Params._dummy(), "removeCols", "cols to remove")
    convertToArrayFirst = Param(Params._dummy(), "convertToArrayFirst", "convert to array first")
    weights = Param(Params._dummy(), "weights", "weights")
    dimNum = Param(Params._dummy(), "dimNum", "dim num")

    def __init__(
        self,
        task_name: str = None,
        input_cols: Optional[List[str]] = None,
        output_col: str = "averaged_values",
        remove_cols: Optional[List[str]] = None,
        convert_to_array_first: bool = False,
        weights: Optional[List[int]] = None,
        dim_num: int = 1,
    ):
        """
        Args:
            task_name (str, optional): Task name: "binary", "multiclass" or "reg".
            input_cols (List[str], optional): List of input columns.
            output_col (str, optional): Output column name. Defaults to "averaged_values".
            remove_cols (Optional[List[str]], optional): Columns need to remove. Defaults to None.
            convert_to_array_first (bool, optional): If `True` then will be convert input vectors to arrays.
                Defaults to False.
            weights (Optional[List[int]], optional): List of weights to scaling output values. Defaults to None.
            dim_num (int, optional): Dimension of input columns. Defaults to 1.
        """
        super().__init__()
        input_cols = input_cols if input_cols else []
        self.set(self.taskName, task_name)
        self.set(self.inputCols, input_cols)
        self.set(self.outputCol, output_col)
        if not remove_cols:
            remove_cols = []
        self.set(self.removeCols, remove_cols)
        self.set(self.convertToArrayFirst, convert_to_array_first)
        if weights is None:
            weights = [1.0 for _ in input_cols]

        assert len(input_cols) == len(weights)

        self.set(self.weights, weights)
        self.set(self.dimNum, dim_num)

    def get_task_name(self) -> str:
        return self.getOrDefault(self.taskName)

    def get_remove_cols(self) -> List[str]:
        return self.getOrDefault(self.removeCols)

    def get_convert_to_array_first(self) -> bool:
        return self.getOrDefault(self.convertToArrayFirst)

    def get_weights(self) -> List[int]:
        return self.getOrDefault(self.weights)

    def get_dim_num(self) -> int:
        return self.getOrDefault(self.dimNum)

    def _transform(self, dataset: SparkDataFrame) -> SparkDataFrame:
        logger.debug(f"In {type(self)}. Columns: {sorted(dataset.columns)}")

        pred_cols = self.getInputCols()

        if self.get_task_name() in ["binary", "multiclass"]:
            out_col = vector_averaging(sf.array(*pred_cols), sf.lit(self.get_dim_num()))
        else:
            out_col = scalar_averaging(sf.array(*pred_cols))

        cols_to_remove = set(self.get_remove_cols())
        cols_to_select = [c for c in dataset.columns if c not in cols_to_remove]
        out_df = dataset.select(*cols_to_select, out_col.alias(self.getOutputCol()))

        logger.debug(f"Out {type(self)}. Columns: {sorted(out_df.columns)}")

        return out_df



import functools
import logging
import random
import time
import warnings

from copy import copy
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union
from typing import cast

import lightgbm as lgb
import pandas as pd
import pyspark.sql.functions as sf

from lightautoml.ml_algo.tuning.base import Distribution
from lightautoml.ml_algo.tuning.base import SearchSpace
from lightautoml.pipelines.selection.base import ImportanceEstimator
from lightautoml.utils.timer import TaskTimer
from lightautoml.validation.base import TrainValidIterator
from lightgbm import Booster
from pandas import Series
from pyspark.ml import PipelineModel
from pyspark.ml import Transformer
from pyspark.ml.feature import VectorAssembler
from pyspark.ml.util import MLReadable
from pyspark.ml.util import MLWritable
from pyspark.ml.util import MLWriter
from synapse.ml.lightgbm import LightGBMClassificationModel
from synapse.ml.lightgbm import LightGBMClassifier
from synapse.ml.lightgbm import LightGBMRegressionModel
from synapse.ml.lightgbm import LightGBMRegressor
from synapse.ml.onnx import ONNXModel

from sparklightautoml.dataset.base import SparkDataset
from sparklightautoml.dataset.roles import NumericVectorOrArrayRole
from sparklightautoml.ml_algo.base import AveragingTransformer
from sparklightautoml.ml_algo.base import ComputationalParameters
from sparklightautoml.ml_algo.base import SparkMLModel
from sparklightautoml.ml_algo.base import SparkTabularMLAlgo
from sparklightautoml.mlwriters import LightGBMModelWrapperMLReader
from sparklightautoml.mlwriters import LightGBMModelWrapperMLWriter
from sparklightautoml.mlwriters import ONNXModelWrapperMLReader
from sparklightautoml.mlwriters import ONNXModelWrapperMLWriter
from sparklightautoml.transformers.base import DropColumnsTransformer
from sparklightautoml.transformers.base import PredictionColsTransformer
from sparklightautoml.transformers.base import ProbabilityColsTransformer
from sparklightautoml.utils import SparkDataFrame
from sparklightautoml.validation.base import SparkBaseTrainValidIterator
from sparklightautoml.validation.base import split_out_val


logger = logging.getLogger(__name__)


class LightGBMModelWrapper(Transformer, MLWritable, MLReadable):
    """Simple wrapper for `synapse.ml.lightgbm.[LightGBMRegressionModel|LightGBMClassificationModel]`
    to fix issue with loading model from saved composite pipeline.

    For more details see: https://github.com/microsoft/SynapseML/issues/614.
    """

    def __init__(self, model: Union[LightGBMRegressionModel, LightGBMClassificationModel] = None) -> None:
        super().__init__()
        self.model = model

    def write(self) -> MLWriter:
        return LightGBMModelWrapperMLWriter(self)

    @classmethod
    def read(cls):
        """Returns an MLReader instance for this class."""
        return LightGBMModelWrapperMLReader()

    def _transform(self, dataset: SparkDataFrame) -> SparkDataFrame:
        return self.model.transform(dataset)


class ONNXModelWrapper(Transformer, MLWritable, MLReadable):
    """Simple wrapper for `ONNXModel` to fix issue with loading model from saved composite pipeline.

    For more details see: https://github.com/microsoft/SynapseML/issues/614.
    """

    def __init__(self, model: ONNXModel = None) -> None:
        super().__init__()
        self.model = model

    def write(self) -> MLWriter:
        return ONNXModelWrapperMLWriter(self)

    @classmethod
    def read(cls):
        """Returns an MLReader instance for this class."""
        return ONNXModelWrapperMLReader()

    def _transform(self, dataset: SparkDataFrame) -> SparkDataFrame:
        return self.model.transform(dataset)


class SparkBoostLGBM(SparkTabularMLAlgo, ImportanceEstimator):
    """Gradient boosting on decision trees from LightGBM library.

    default_params: All available parameters listed in synapse.ml documentation:

    - https://mmlspark.blob.core.windows.net/docs/0.9.5/pyspark/synapse.ml.lightgbm.html#module-synapse.ml.lightgbm.LightGBMClassifier
    - https://mmlspark.blob.core.windows.net/docs/0.9.5/pyspark/synapse.ml.lightgbm.html#module-synapse.ml.lightgbm.LightGBMRegressor

    freeze_defaults:

        - ``True`` :  params may be rewritten depending on dataset.
        - ``False``:  params may be changed only manually or with tuning.

    timer: :class:`~lightautoml.utils.timer.Timer` instance or ``None``.

    """

    _name: str = "LightGBM"

    _default_params = {
        "learningRate": 0.05,
        "numLeaves": 128,
        "featureFraction": 0.7,
        "baggingFraction": 0.7,
        "baggingFreq": 1,
        "maxDepth": -1,
        "minGainToSplit": 0.0,
        "maxBin": 255,
        "minDataInLeaf": 5,
        # e.g. num trees
        "numIterations": 3000,
        "earlyStoppingRound": 50,
        # for regression
        "alpha": 1.0,
        "lambdaL1": 0.0,
        "lambdaL2": 0.0,
    }

    # mapping between metric name defined via SparkTask
    # and metric names supported by LightGBM
    _metric2lgbm = {
        "binary": {"auc": "auc", "aupr": "areaUnderPR"},
        "reg": {
            "r2": "rmse",
            "mse": "mse",
            "mae": "mae",
        },
        "multiclass": {"crossentropy": "cross_entropy"},
    }

    def __init__(
        self,
        default_params: Optional[dict] = None,
        freeze_defaults: bool = True,
        timer: Optional[TaskTimer] = None,
        optimization_search_space: Optional[dict] = None,
        use_single_dataset_mode: bool = True,
        max_validation_size: int = 10_000,
        chunk_size: int = 4_000_000,
        convert_to_onnx: bool = False,
        mini_batch_size: int = 5000,
        seed: int = 42,
        parallelism: int = 1,
        execution_mode: str = "bulk",
        use_barrier_execution_mode: bool = False,
        experimental_parallel_mode: bool = False,
        persist_output_dataset: bool = True,
        computations_settings: Optional[ComputationalParameters] = None,
    ):
        optimization_search_space = optimization_search_space if optimization_search_space else dict()
        SparkTabularMLAlgo.__init__(
            self,
            default_params,
            freeze_defaults,
            timer,
            optimization_search_space,
            persist_output_dataset,
            computations_settings,
        )
        self._probability_col_name = "probability"
        self._prediction_col_name = "prediction"
        self._raw_prediction_col_name = "raw_prediction"
        self._assembler = None
        self._drop_cols_transformer = None
        self._use_single_dataset_mode = use_single_dataset_mode
        self._max_validation_size = max_validation_size
        self._seed = seed
        self._models_feature_importances = []
        self._chunk_size = chunk_size
        self._convert_to_onnx = convert_to_onnx
        self._mini_batch_size = mini_batch_size
        self._parallelism = parallelism
        self._executin_mode = execution_mode
        self._use_barrier_execution_mode = use_barrier_execution_mode
        self._experimental_parallel_mode = experimental_parallel_mode

    def _infer_params(self, runtime_settings: Optional[Dict[str, Any]] = None) -> Tuple[dict, int]:
        """Infer all parameters in lightgbm format.

        Returns:
            Tuple (params, num_trees, early_stopping_rounds, verbose_eval, fobj, feval).
            About parameters: https://lightgbm.readthedocs.io/en/latest/_modules/lightgbm/engine.html

        """
        assert self.task is not None

        # TODO: PARALLEL - validate runtime settings

        task = self.task.name

        params = copy(self.params)

        if "isUnbalance" in params:
            params["isUnbalance"] = True if params["isUnbalance"] == 1 else False

        verbose_eval = 1

        if task == "reg":
            params["objective"] = "regression"
            params["metric"] = self._metric2lgbm[task].get(self.task.metric_name, None)
        elif task == "binary":
            params["objective"] = "binary"
            params["metric"] = self._metric2lgbm[task].get(self.task.metric_name, None)
        elif task == "multiclass":
            params["objective"] = "multiclass"
            params["metric"] = "multiclass"
        else:
            raise ValueError(f"Unsupported task type: {task}")

        if task != "reg":
            if "alpha" in params:
                del params["alpha"]
            if "lambdaL1" in params:
                del params["lambdaL1"]
            if "lambdaL2" in params:
                del params["lambdaL2"]

        runtime_settings = runtime_settings or dict()
        if "num_tasks" in runtime_settings:
            params["numTasks"] = runtime_settings["num_tasks"]
        if "num_threads" in runtime_settings:
            params["numThreads"] = runtime_settings["num_threads"]

        return params, verbose_eval

    def init_params_on_input(self, train_valid_iterator: TrainValidIterator) -> dict:
        self.task = train_valid_iterator.train.task

        sds = cast(SparkDataset, train_valid_iterator.train)
        rows_num = sds.data.count()
        task = train_valid_iterator.train.task.name

        suggested_params = copy(self.default_params)

        if self.freeze_defaults:
            # if user change defaults manually - keep it
            return suggested_params

        if task == "reg":
            suggested_params = {
                "learningRate": 0.05,
                "numLeaves": 32,
                "featureFraction": 0.9,
                "baggingFraction": 0.9,
            }

        if rows_num <= 10000:
            init_lr = 0.01
            ntrees = 3000
            es = 200

        elif rows_num <= 20000:
            init_lr = 0.02
            ntrees = 3000
            es = 200

        elif rows_num <= 100000:
            init_lr = 0.03
            ntrees = 2000
            es = 200
        elif rows_num <= 300000:
            init_lr = 0.04
            ntrees = 2000
            es = 100
        else:
            init_lr = 0.05
            ntrees = 2000
            es = 100

        if rows_num > 300000:
            suggested_params["numLeaves"] = 128 if task == "reg" else 244
        elif rows_num > 100000:
            suggested_params["numLeaves"] = 64 if task == "reg" else 128
        elif rows_num > 50000:
            suggested_params["numLeaves"] = 32 if task == "reg" else 64
            # params['reg_alpha'] = 1 if task == 'reg' else 0.5
        elif rows_num > 20000:
            suggested_params["numLeaves"] = 32 if task == "reg" else 32
            suggested_params["alpha"] = 0.5 if task == "reg" else 0.0
        elif rows_num > 10000:
            suggested_params["numLeaves"] = 32 if task == "reg" else 64
            suggested_params["alpha"] = 0.5 if task == "reg" else 0.2
        elif rows_num > 5000:
            suggested_params["numLeaves"] = 24 if task == "reg" else 32
            suggested_params["alpha"] = 0.5 if task == "reg" else 0.5
        else:
            suggested_params["numLeaves"] = 16 if task == "reg" else 16
            suggested_params["alpha"] = 1 if task == "reg" else 1

        suggested_params["learningRate"] = init_lr
        suggested_params["numIterations"] = ntrees
        suggested_params["earlyStoppingRound"] = es

        if task != "reg":
            if "alpha" in suggested_params:
                del suggested_params["alpha"]

        return suggested_params

    def _get_default_search_spaces(self, suggested_params: Dict, estimated_n_trials: int) -> Dict:
        """Train on train dataset and predict on holdout dataset.

        Args:.
            suggested_params: suggested params
            estimated_n_trials: Number of trials.

        Returns:
            Target predictions for valid dataset.

        """
        assert self.task is not None

        optimization_search_space = dict()

        optimization_search_space["featureFraction"] = SearchSpace(
            Distribution.UNIFORM,
            low=0.5,
            high=1.0,
        )

        optimization_search_space["numLeaves"] = SearchSpace(
            Distribution.INTUNIFORM,
            low=4,
            high=255,
        )

        if self.task.name == "binary" or self.task.name == "multiclass":
            optimization_search_space["isUnbalance"] = SearchSpace(Distribution.DISCRETEUNIFORM, low=0, high=1, q=1)

        if estimated_n_trials > 30:
            optimization_search_space["baggingFraction"] = SearchSpace(
                Distribution.UNIFORM,
                low=0.5,
                high=1.0,
            )

            optimization_search_space["minSumHessianInLeaf"] = SearchSpace(
                Distribution.LOGUNIFORM,
                low=1e-3,
                high=10.0,
            )

        if estimated_n_trials > 100:
            if self.task.name == "reg":
                optimization_search_space["alpha"] = SearchSpace(
                    Distribution.LOGUNIFORM,
                    low=1e-8,
                    high=10.0,
                )

            optimization_search_space["lambdaL1"] = SearchSpace(
                Distribution.LOGUNIFORM,
                low=1e-8,
                high=10.0,
            )

        return optimization_search_space

    def predict_single_fold(self, dataset: SparkDataset, model: PipelineModel) -> SparkDataFrame:
        return model.transform(dataset.data)

    def _do_convert_to_onnx(self, train: SparkDataset, ml_model):
        logger.info("Model convert is started")
        booster_model_str = ml_model.getLightGBMBooster().modelStr().get()
        booster = lgb.Booster(model_str=booster_model_str)
        model_payload_ml = self._convert_model(booster, len(train.features))

        onnx_ml = ONNXModel().setModelPayload(model_payload_ml)

        if train.task.name == "reg":
            onnx_ml = (
                onnx_ml.setDeviceType("CPU")
                .setFeedDict({"input": f"{self._name}_vassembler_features"})
                .setFetchDict({ml_model.getPredictionCol(): "variable"})
                .setMiniBatchSize(self._mini_batch_size)
            )
        else:
            onnx_ml = (
                onnx_ml.setDeviceType("CPU")
                .setFeedDict({"input": f"{self._name}_vassembler_features"})
                .setFetchDict({ml_model.getProbabilityCol(): "probabilities", ml_model.getPredictionCol(): "label"})
                .setMiniBatchSize(self._mini_batch_size)
            )

        logger.info("Model convert is ended")

        return onnx_ml

    def fit_predict_single_fold(
        self,
        fold_prediction_column: str,
        validation_column: str,
        train: SparkDataset,
        runtime_settings: Optional[Dict[str, Any]] = None,
    ) -> Tuple[SparkMLModel, SparkDataFrame, str]:
        if self.task is None:
            self.task = train.task

        (params, verbose_eval) = self._infer_params(runtime_settings)

        logger.info(f"Input cols for the vector assembler: {train.features}")
        logger.info(f"Running lgb with the following params: {params}")

        if train.task.name in ["binary", "multiclass"]:
            params["rawPredictionCol"] = self._raw_prediction_col_name
            params["probabilityCol"] = fold_prediction_column
            params["predictionCol"] = self._prediction_col_name
            params["isUnbalance"] = True
        else:
            params["predictionCol"] = fold_prediction_column

        assert (
            validation_column in train.data.columns
        ), f"Validation column {validation_column} should be present in the data"

        full_data = self._ensure_validation_size(train.data, validation_column)

        # prepare assembler
        if self._assembler is None:
            self._assembler = VectorAssembler(
                inputCols=train.features, outputCol=f"{self._name}_vassembler_features", handleInvalid="keep"
            )

        # assign a random port to decrease chances of allocating the same port from multiple instances
        rand = random.Random(time.time_ns())
        random_port = rand.randint(10_000, 50_000)

        run_params = {
            "featuresCol": self._assembler.getOutputCol(),
            "labelCol": train.target_column,
            "validationIndicatorCol": validation_column,
            "verbosity": verbose_eval,
            "executionMode": self._executin_mode,
            "useSingleDatasetMode": self._use_single_dataset_mode,
            "useBarrierExecutionMode": self._use_barrier_execution_mode,
            "isProvideTrainingMetric": True,
            "chunkSize": self._chunk_size,
            "defaultListenPort": random_port,
            **params,
            **({"alpha": 0.5, "lambdaL1": 0.0, "lambdaL2": 0.0} if train.task.name == "reg" else dict()),
        }

        # build the booster
        lgbm_booster = LightGBMRegressor if train.task.name == "reg" else LightGBMClassifier
        lgbm = lgbm_booster(**run_params)

        logger.info(f"Use single dataset mode: {lgbm.getUseSingleDatasetMode()}. NumThreads: {lgbm.getNumThreads()}")
        logger.info(f"All lgbm booster params: {run_params}")

        rows_count = full_data.count()
        if (run_params["executionMode"] == "streaming") and (rows_count <= 25_000):
            warnings.warn(
                f"The fitting of lightgbm in streaming execution mode "
                f"may fail with SEGSIGV / SIGBUS error (probably due to a bug in synapse ml) "
                f"if too few data available per core. Train data rows count: {rows_count} "
                f"Consider switching to bulk execution mode if such crashes happen",
                RuntimeWarning,
            )

        # fitting the model
        ml_model = lgbm.fit(self._assembler.transform(full_data))

        # handle the model
        ml_model = self._do_convert_to_onnx(train, ml_model) if self._convert_to_onnx else ml_model
        self._models_feature_importances.append(ml_model.getFeatureImportances(importance_type="gain"))

        valid_data = split_out_val(full_data, validation_column)
        # predict validation
        val_pred = ml_model.transform(self._assembler.transform(valid_data))
        val_pred = DropColumnsTransformer(
            remove_cols=[],
            optional_remove_cols=[self._prediction_col_name, self._probability_col_name, self._raw_prediction_col_name],
        ).transform(val_pred)

        return ml_model, val_pred, fold_prediction_column

    def fit(self, train_valid: SparkBaseTrainValidIterator):
        logger.info("Starting LGBM fit")
        self.fit_predict(train_valid)
        logger.info("Finished LGBM fit")

    def get_features_score(self) -> Series:
        imp = functools.reduce(lambda acc, x: acc + pd.Series(x), self._models_feature_importances, 0)

        # imp = 0
        # for model_feature_impotances in self._models_feature_importances:
        #     imp += pd.Series(model_feature_impotances)

        imp /= len(self._models_feature_importances)

        def flatten_features(feat: str):
            role = self.input_roles[feat]
            if isinstance(role, NumericVectorOrArrayRole):
                return [f"{feat}_pos_{i}" for i in range(role.size)]
            return [feat]

        index = [ff for feat in self._assembler.getInputCols() for ff in flatten_features(feat)]

        result = Series(list(imp), index=index).sort_values(ascending=False)
        return result

    @staticmethod
    def _convert_model(lgbm_model: Booster, input_size: int) -> bytes:
        from onnxconverter_common.data_types import FloatTensorType
        from onnxmltools.convert import convert_lightgbm

        initial_types = [("input", FloatTensorType([-1, input_size]))]
        onnx_model = convert_lightgbm(lgbm_model, initial_types=initial_types, target_opset=9)
        return onnx_model.SerializeToString()

    def _ensure_validation_size(self, full_data: SparkDataFrame, validation_column: str) -> SparkDataFrame:
        # reduce validation size if it is too big
        val_data_size = full_data.where(sf.col(validation_column).astype("int") == 1).count()
        if val_data_size > self._max_validation_size:
            logger.warning(
                f"Too big validation fold: {val_data_size}. "
                f"Reducing its size down according to max_validation_size setting:"
                f" {self._max_validation_size}"
            )
            not_val_row = sf.col(validation_column) != sf.lit(1)
            not_chosen_as_val = sf.rand(seed=self._seed) < sf.lit(self._max_validation_size / val_data_size)
            full_data = full_data.where(not_val_row | not_chosen_as_val)

        # checking if there are no empty partitions that may lead to hanging
        rows = (
            full_data.withColumn("__partition_id__", sf.spark_partition_id())
            .groupby("__partition_id__")
            .agg(sf.sum(validation_column).alias("val_values"), sf.count("*").alias("all_values"))
            .collect()
        )
        for row in rows:
            if row["val_values"] == row["all_values"] or row["all_values"] == 0:
                warnings.warn(
                    f"Empty partition encountered: partition id - {row['__partition_id_']},"
                    f"validation values count in the partition - {row['val_values']}, "
                    f"all values count in the partition  - {row['all_values']}"
                )
                raise ValueError(
                    f"Empty partition encountered: partition id - {row['__partition_id_']},"
                    f"validation values count in the partition - {row['val_values']}, "
                    f"all values count in the partition  - {row['all_values']}"
                )

        return full_data

    def _build_transformer(self) -> Transformer:
        avr = self._build_averaging_transformer()
        if self._convert_to_onnx:
            wrapped_models = [ONNXModelWrapper(m) for m in self.models]
        else:
            wrapped_models = [LightGBMModelWrapper(m) for m in self.models]
        models: List[Transformer] = [
            el
            for m in wrapped_models
            for el in [
                m,
                DropColumnsTransformer(
                    remove_cols=[],
                    optional_remove_cols=[
                        self._prediction_col_name,
                        self._probability_col_name,
                        self._raw_prediction_col_name,
                    ],
                ),
            ]
        ]
        if self._convert_to_onnx:
            if self.task.name in ["binary", "multiclass"]:
                models.append(
                    ProbabilityColsTransformer(
                        probability_cols=self._models_prediction_columns, num_classes=self.n_classes
                    )
                )
            else:
                models.append(PredictionColsTransformer(prediction_cols=self._models_prediction_columns))
        averaging_model = PipelineModel(
            stages=[
                self._assembler,
                *models,
                avr,
                self._build_vector_size_hint(self.prediction_feature, self.prediction_role),
            ]
        )
        return averaging_model

    def _build_averaging_transformer(self) -> Transformer:
        avr = AveragingTransformer(
            self.task.name,
            input_cols=self._models_prediction_columns,
            output_col=self.prediction_feature,
            remove_cols=[self._assembler.getOutputCol(), *self._models_prediction_columns],
            convert_to_array_first=not (self.task.name == "reg"),
            dim_num=self.n_classes,
        )
        return avr

    def fit_predict(self, train_valid_iterator: SparkBaseTrainValidIterator) -> SparkDataset:
        """Fit and then predict accordig the strategy that uses train_valid_iterator.

        If item uses more then one time it will
        predict mean value of predictions.
        If the element is not used in training then
        the prediction will be ``numpy.nan`` for this item

        Args:
            train_valid_iterator: Classic cv-iterator.

        Returns:
            Dataset with predicted values.

        """
        logger.info("Starting LGBM fit")
        self.timer.start()

        res = super().fit_predict(train_valid_iterator)

        logger.info("Finished LGBM fit")
        return res



from typing import Any
from typing import Dict
from typing import Optional
from typing import Tuple
from typing import Union

from sparklightautoml.computations.base import ComputationsManager
from sparklightautoml.computations.base import ComputationsSettings
from sparklightautoml.computations.parallel import ParallelComputationsManager
from sparklightautoml.computations.sequential import SequentialComputationsManager


# either named profile and parallelism or parallelism settings or factory
AutoMLComputationsSettings = Union[Tuple[str, int], Dict[str, Any], "ComputationsManagerFactory"]


class ComputationsManagerFactory:
    def __init__(self, computations_settings: Optional[Union[Tuple[str, int], Dict[str, Any]]] = None):
        super(ComputationsManagerFactory, self).__init__()
        computations_settings = computations_settings or ("no_parallelism", -1)

        if isinstance(computations_settings, Tuple):
            mode, parallelism = computations_settings
            self._computations_settings = build_named_parallelism_settings(mode, parallelism)
        else:
            self._computations_settings = computations_settings

        self._ml_pipelines_params = self._computations_settings.get("ml_pipelines", None)
        self._ml_algos_params = self._computations_settings.get("ml_algos", None)
        self._selector_params = self._computations_settings.get("selector", None)
        self._tuner_params = self._computations_settings.get("tuner", None)
        self._linear_l2_params = self._computations_settings.get("linear_l2", None)
        self._lgb_params = self._computations_settings.get("lgb", None)

    def get_ml_pipelines_manager(self) -> "ComputationsManager":
        return build_computations_manager(self._ml_pipelines_params)

    def get_ml_algo_manager(self) -> "ComputationsManager":
        return build_computations_manager(self._ml_algos_params)

    def get_selector_manager(self) -> "ComputationsManager":
        return build_computations_manager(self._selector_params)

    def get_tuning_manager(self) -> "ComputationsManager":
        return build_computations_manager(self._tuner_params)

    def get_lgb_manager(self) -> "ComputationsManager":
        return build_computations_manager(self._lgb_params)

    def get_linear_manager(self) -> "ComputationsManager":
        return build_computations_manager(self._linear_l2_params)


def build_named_parallelism_settings(config_name: str, parallelism: int):
    intra_mlpipe_parallelim = {
        "ml_pipelines": {"parallelism": 1},
        "ml_algos": {"parallelism": 1},
        "selector": {"parallelism": parallelism},
        "tuner": {"parallelism": parallelism},
        "linear_l2": {"parallelism": parallelism},
        "lgb": {"parallelism": parallelism, "use_location_prefs_mode": False},
    }

    parallelism_config = {
        "no_parallelism": {},
        "parallelism": intra_mlpipe_parallelim,
        "intra_mlpipe_parallelism": intra_mlpipe_parallelim,
        "intra_mlpipe_parallelism_with_location_prefs_mode": {
            "ml_pipelines": {"parallelism": 1},
            "ml_algos": {"parallelism": 1},
            "selector": {"parallelism": parallelism},
            "tuner": {"parallelism": parallelism},
            "linear_l2": {"parallelism": parallelism},
            "lgb": {"parallelism": parallelism, "use_location_prefs_mode": True},
        },
        "mlpipe_level_parallelism": {
            "ml_pipelines": {"parallelism": parallelism},
            "ml_algos": {"parallelism": 1},
            "selector": {"parallelism": 1},
            "tuner": {"parallelism": 1},
            "linear_l2": {"parallelism": 1},
            "lgb": {"parallelism": 1},
        },
    }

    assert config_name in parallelism_config, (
        f"Not supported parallelism mode: {config_name}. "
        f"Only the following ones are supoorted at the moment: {list(parallelism_config.keys())}"
    )

    return parallelism_config[config_name]


def build_computations_manager(computations_settings: Optional[ComputationsSettings] = None) -> "ComputationsManager":
    if computations_settings is not None and isinstance(computations_settings, ComputationsManager):
        computations_manager = computations_settings
    elif computations_settings is not None:
        assert isinstance(computations_settings, dict)
        parallelism = int(computations_settings.get("parallelism", "1"))
        use_location_prefs_mode = computations_settings.get("use_location_prefs_mode", False)
        computations_manager = ParallelComputationsManager(parallelism, use_location_prefs_mode)
    else:
        computations_manager = SequentialComputationsManager()

    return computations_manager



"""Base AutoML class."""
import functools
import logging
import os

from copy import copy
from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union
from typing import cast

from lightautoml.reader.base import RolesDict
from lightautoml.utils.logging import set_stdout_level
from lightautoml.utils.logging import verbosity_to_loglevel
from lightautoml.utils.timer import PipelineTimer
from pyspark.ml import PipelineModel
from pyspark.ml import Transformer

from ..computations.base import ComputationsManager
from ..computations.base import ComputationsSettings
from ..computations.builder import build_computations_manager
from ..dataset.base import PersistenceLevel
from ..dataset.base import PersistenceManager
from ..dataset.base import SparkDataset
from ..dataset.persistence import PlainCachePersistenceManager
from ..pipelines.base import TransformerInputOutputRoles
from ..pipelines.features.base import SparkPipelineModel
from ..pipelines.ml.base import SparkMLPipeline
from ..reader.base import SparkToSparkReader
from ..utils import ColumnsSelectorTransformer
from ..utils import SparkDataFrame
from ..utils import get_current_session
from ..validation.base import SparkBaseTrainValidIterator
from ..validation.base import mark_as_train
from ..validation.base import mark_as_val
from ..validation.iterators import SparkDummyIterator
from ..validation.iterators import SparkFoldsIterator
from ..validation.iterators import SparkHoldoutIterator
from .blend import SparkBestModelSelector
from .blend import SparkBlender


logger = logging.getLogger(__name__)

# Either path/full url, or pyspark.sql.DataFrame
ReadableIntoSparkDf = Union[str, SparkDataFrame]


class SparkAutoML(TransformerInputOutputRoles):
    """Class for compile full pipeline of AutoML task.

    AutoML steps:

        - Read, analyze data and get inner
          :class:`~lightautoml.dataset.base.LAMLDataset` from input
          dataset: performed by reader.
        - Create validation scheme.
        - Compute passed ml pipelines from levels.
          Each element of levels is list
          of :class:`~lightautoml.pipelines.ml.base.MLPipelines`
          prediction from current level are passed to next level
          pipelines as features.
        - Time monitoring - check if we have enough time to calc new pipeline.
        - Blend last level models and prune useless pipelines
          to speedup inference: performed by blender.
        - Returns prediction on validation data.
          If crossvalidation scheme is used,
          out-of-fold prediction will returned.
          If validation data is passed
          it will return prediction on validation dataset.
          In case of cv scheme when some point of train data
          never was used as validation (ex. timeout exceeded
          or custom cv iterator like
          :class:`~lightautoml.validation.np_iterators.TimeSeriesIterator`
          was used) NaN for this point will be returned.

    Example:
        Common usecase - create custom pipelines or presets.

        >>> reader = SparkToSparkReader()
        >>> pipe = SparkMLPipeline([SparkMLAlgo()])
        >>> levels = [[pipe]]
        >>> automl = SparkAutoML(reader, levels, )
        >>> automl.fit_predict(data, roles={'target': 'TARGET'})

    """

    def __init__(
        self,
        reader: Optional[SparkToSparkReader] = None,
        levels: Optional[Sequence[Sequence[SparkMLPipeline]]] = None,
        timer: Optional[PipelineTimer] = None,
        blender: Optional[SparkBlender] = None,
        skip_conn: bool = False,
        return_all_predictions: bool = False,
        computation_settings: Optional[ComputationsSettings] = ("no_parallelism", -1),
    ):
        """

        Args:
            reader: Instance of Reader class object that
              creates :class:`~lightautoml.dataset.base.LAMLDataset`
              from input data.
            levels: List of list
              of :class:`~lightautoml.pipelines.ml..base.MLPipelines`.
            timer: Timer instance of
              :class:`~lightautoml.utils.timer.PipelineTimer`.
              Default - unlimited timer.
            blender: Instance of Blender.
              Default - :class:`~lightautoml.automl.blend.BestModelSelector`.
            skip_conn: True if we should pass first level
              input features to next levels.

        Note:
            There are several verbosity levels:

                - `0`: No messages.
                - `1`: Warnings.
                - `2`: Info.
                - `3`: Debug.

        """
        super().__init__()
        self.levels: Optional[Sequence[Sequence[SparkMLPipeline]]] = None
        self._transformer = None
        self._input_roles: Optional[RolesDict] = None
        self._output_roles: Optional[RolesDict] = None
        self._service_columns: Optional[List[str]] = None
        self._persistence_manager: Optional[PersistenceManager] = None
        if reader and levels:
            self._initialize(reader, levels, timer, blender, skip_conn, return_all_predictions)

        self._computations_manager: Optional[ComputationsManager] = build_computations_manager(computation_settings)

    @property
    def input_roles(self) -> Optional[RolesDict]:
        return self._input_roles

    @property
    def output_roles(self) -> Optional[RolesDict]:
        return self._output_roles

    @property
    def persistence_manager(self) -> Optional[PersistenceManager]:
        return self._persistence_manager

    def transformer(
        self,
        return_all_predictions: bool = False,
        add_array_attrs: bool = True,
        leave_only_predict_cols: bool = False,
        **reader_args,
    ) -> SparkPipelineModel:
        if not return_all_predictions:
            blender = [self.blender.transformer()]
            output_roles = self.blender.output_roles
        else:
            blender = []
            output_roles = {**(ml_pipe.output_roles for ml_pipe in self.levels[-1])}

        sel_tr = ColumnsSelectorTransformer(
            name="SparkAutoML",
            input_cols=[SparkDataset.ID_COLUMN] + list(output_roles.keys()),
            optional_cols=[self.reader.target_col] if self.reader.target_col else [],
        )

        stages = [
            self.reader.transformer(add_array_attrs=add_array_attrs, **reader_args),
            *(ml_pipe.transformer() for level in self.levels for ml_pipe in level),
            *blender,
            *([sel_tr] if leave_only_predict_cols else []),
        ]

        return SparkPipelineModel(stages, input_roles=self.input_roles, output_roles=output_roles)

    def _initialize(
        self,
        reader: SparkToSparkReader,
        levels: Sequence[Sequence[SparkMLPipeline]],
        timer: Optional[PipelineTimer] = None,
        blender: Optional[SparkBlender] = None,
        skip_conn: bool = False,
        return_all_predictions: bool = False,
    ):
        """Same as __init__. Exists for delayed initialization in presets.

        Args:
            reader: Instance of Reader class object that
              creates :class:`~lightautoml.dataset.base.LAMLDataset`
              from input data.
            levels: List of list
              of :class:`~lightautoml.pipelines.ml..base.MLPipelines`.
            timer: Timer instance of
              :class:`~lightautoml.utils.timer.PipelineTimer`.
              Default - unlimited timer.
            blender: Instance of Blender.
              Default - :class:`~lightautoml.automl.blend.BestModelSelector`.
            skip_conn: True if we should pass first level
              input features to next levels.
            return_all_predictions: True if we should return all predictions from last
              level models.

        """
        assert len(levels) > 0, "At least 1 level should be defined"

        self.timer = timer
        if timer is None:
            self.timer = PipelineTimer()
        self.reader = reader
        self._levels = levels

        # default blender is - select best model and prune other pipes
        self.blender = blender
        if blender is None:
            self.blender = SparkBestModelSelector()

        # update model names
        for i, lvl in enumerate(self._levels):
            for j, pipe in enumerate(lvl):
                pipe.upd_model_names("Lvl_{0}_Pipe_{1}".format(i, j))

        self.skip_conn = skip_conn
        self.return_all_predictions = return_all_predictions

    def fit_predict(
        self,
        train_data: Any,
        roles: dict,
        train_features: Optional[Sequence[str]] = None,
        cv_iter: Optional[Iterable] = None,
        valid_data: Optional[Any] = None,
        valid_features: Optional[Sequence[str]] = None,
        verbose: int = 0,
        persistence_manager: Optional[PersistenceManager] = None,
    ) -> SparkDataset:
        """Fit on input data and make prediction on validation part.

        Args:
            train_data: Dataset to train.
            roles: Roles dict.
            train_features: Optional features names,
              if cannot be inferred from train_data.
            cv_iter: Custom cv iterator. For example,
              :class:`~lightautoml.validation.np_iterators.TimeSeriesIterator`.
            valid_data: Optional validation dataset.
            valid_features: Optional validation dataset
              features if can't be inferred from `valid_data`.
            verbose: controls verbosity

        Returns:
            Predicted values.

        """
        set_stdout_level(verbosity_to_loglevel(verbose))
        self.timer.start()

        self._persistence_manager = persistence_manager or PlainCachePersistenceManager()

        train_dataset = self.reader.fit_read(
            train_data, train_features, roles, persistence_manager=self._persistence_manager
        )

        train_dataset = train_dataset.persist(level=PersistenceLevel.REGULAR)

        assert (
            len(self._levels) <= 1 or train_dataset.folds_column is not None
        ), "Not possible to fit more than 1 level without cv folds"

        assert (
            len(self._levels) <= 1 or valid_data is None
        ), "Not possible to fit more than 1 level with holdout validation"

        if valid_data:
            valid_dataset = self.reader.read(valid_data, valid_features, add_array_attrs=True).persist(
                PersistenceLevel.READER
            )
        else:
            valid_dataset = None

        train_valid = self._create_validation_iterator(train_dataset, valid_dataset, None, cv_iter=cv_iter)

        pipes: List[SparkMLPipeline] = []
        self.levels = []
        level_ds: Optional[SparkDataset] = None
        for leven_number, level in enumerate(self._levels, 1):
            pipes = []
            flg_last_level = leven_number == len(self._levels)

            logger.info(
                f"Layer \x1b[1m{leven_number}\x1b[0m train process start. Time left {self.timer.time_left:.2f} secs"
            )

            with train_valid.frozen() as frozen_train_valid:
                pipes, all_pipes_predictions, flg_last_level = self._parallel_level(
                    level=level, train_valid_iterator=frozen_train_valid
                )

            flg_last_level = flg_last_level or leven_number == len(self._levels)

            logger.info("\x1b[1mLayer {} training completed.\x1b[0m\n".format(leven_number))

            level_ds_name = f"all_piped_predictions_level_{leven_number}"

            if flg_last_level:
                level_ds = SparkDataset.concatenate(all_pipes_predictions, name=level_ds_name)
                train_valid.unpersist()
                break

            self.levels.append(pipes)

            level = (
                [train_valid.get_validation_data(), *all_pipes_predictions] if self.skip_conn else all_pipes_predictions
            )
            name = f"{level_ds_name}_skip_conn" if self.skip_conn else level_ds_name
            level_ds = SparkDataset.concatenate(level, name=name)

            train_valid.unpersist(skip_val=self.skip_conn)
            train_valid = self._create_validation_iterator(level_ds, None, n_folds=None, cv_iter=None)

        blended_prediction, last_pipes = self.blender.fit_predict(level_ds, pipes)
        self.levels.append(last_pipes)

        del self._levels

        oof_pred = level_ds if self.return_all_predictions else blended_prediction

        self._input_roles = copy(train_dataset.roles)
        self._output_roles = copy(oof_pred.roles)
        self._service_columns = train_dataset.service_columns

        return oof_pred

    def predict(
        self,
        data: ReadableIntoSparkDf,
        return_all_predictions: Optional[bool] = None,
        add_reader_attrs: bool = False,
        persistence_manager: Optional[PersistenceManager] = None,
    ) -> SparkDataset:
        """Get dataset with predictions.

        Almost same as :meth:`lightautoml.automl.base.AutoML.predict`
        on new dataset, with additional features.

        Additional features - working with different data formats.
        Supported now:

            - Path to ``.csv``, ``.parquet``, ``.feather`` files.
            - :class:`~numpy.ndarray`, or dict of :class:`~numpy.ndarray`. For example,
              ``{'data': X...}``. In this case roles are optional,
              but `train_features` and `valid_features` required.
            - :class:`pandas.DataFrame`.

        Parallel inference - you can pass ``n_jobs`` to speedup
        prediction (requires more RAM).
        Batch_inference - you can pass ``batch_size``
        to decrease RAM usage (may be longer).

        Args:
            data: Dataset to perform inference.
            features_names: Optional features names,
              if cannot be inferred from `train_data`.
            return_all_predictions: if True,
              returns all model predictions from last level
            add_reader_attrs: if True,
              the reader's attributes will be added to the SparkDataset

        Returns:
            Dataset with predictions.

        """
        persistence_manager = persistence_manager or PlainCachePersistenceManager()
        automl_model = self.transformer(
            return_all_predictions=return_all_predictions,
            add_array_attrs=add_reader_attrs,
            leave_only_predict_cols=True,
        )

        data = self._read_data(data)
        predictions = automl_model.transform(data)

        sds = SparkDataset(
            data=predictions,
            roles=copy(automl_model.get_output_roles()),
            task=self.reader.task,
            persistence_manager=persistence_manager,
            target=self.reader.target_col,
        )

        return sds

    def collect_used_feats(self) -> List[str]:
        """Get feats that automl uses on inference.

        Returns:
            Features names list.

        """
        used_feats = set()

        for lvl in self.levels:
            for pipe in lvl:
                used_feats.update(pipe.used_features)

        used_feats = list(used_feats)

        return used_feats

    def collect_model_stats(self) -> Dict[str, int]:
        """Collect info about models in automl.

        Returns:
            Dict with models and its runtime numbers.

        """
        model_stats = {}

        for lvl in self.levels:
            for pipe in lvl:
                for ml_algo in pipe.ml_algos:
                    model_stats[ml_algo.name] = len(ml_algo.models)

        return model_stats

    def _create_validation_iterator(
        self, train: SparkDataset, valid: Optional[SparkDataset], n_folds: Optional[int], cv_iter: Optional[Callable]
    ) -> SparkBaseTrainValidIterator:
        if valid:
            sdf = mark_as_train(train.data, SparkBaseTrainValidIterator.TRAIN_VAL_COLUMN).unionByName(
                mark_as_val(valid.data, SparkBaseTrainValidIterator.TRAIN_VAL_COLUMN)
            )

            new_train = train.empty()
            new_train.set_data(sdf, train.features, train.roles, dependencies=[train, valid])
            new_train = new_train.persist(level=PersistenceLevel.REGULAR)

            iterator = SparkHoldoutIterator(new_train)
        elif cv_iter:
            raise NotImplementedError("Not supported now")
        elif train.folds_column:
            train = train.persist(level=PersistenceLevel.REGULAR)
            iterator = SparkFoldsIterator(train, n_folds)
        else:
            train = train.persist(level=PersistenceLevel.REGULAR)
            iterator = SparkDummyIterator(train)

        logger.info(f"Using train valid iterator of type: {type(iterator)}")

        return iterator

    def _get_service_columns(self) -> List[str]:
        return self._service_columns

    def _build_transformer(
        self, no_reader: bool = False, return_all_predictions: bool = False
    ) -> Tuple[Transformer, RolesDict]:
        stages = []
        if not no_reader:
            stages.append(self.reader.transformer(add_array_attrs=True))

        ml_pipes = [ml_pipe.transformer() for level in self.levels for ml_pipe in level]
        stages.extend(ml_pipes)

        if not return_all_predictions:
            stages.append(self.blender.transformer())
            output_roles = self.blender.output_roles
        else:
            output_roles = dict()
            for ml_pipe in self.levels[-1]:
                output_roles.update(ml_pipe.output_roles)

        sel_tr = ColumnsSelectorTransformer(
            name="SparkAutoML",
            input_cols=[SparkDataset.ID_COLUMN] + list(output_roles.keys()),
            optional_cols=[self.reader.target_col] if self.reader.target_col else [],
        )
        stages.append(sel_tr)

        automl_transformer = PipelineModel(stages=stages)

        return automl_transformer, output_roles

    def _read_data(self, data: ReadableIntoSparkDf, features: Optional[List[str]] = None) -> SparkDataFrame:
        """Get :class:`~pyspark.sql.DataFrame` from different data formats.

        Note:
            Supported now data formats:

                - Path to ``.csv``, ``.parquet``, ``.feather`` files.
                - :class:`~numpy.ndarray`, or dict of :class:`~numpy.ndarray`.
                  For example, ``{'data': X...}``. In this case,
                  roles are optional, but `train_features`
                  and `valid_features` required.
                - :class:`pandas.DataFrame`.

        Args:
            data: Readable to DataFrame data.
            features_names: Optional features names if ``numpy.ndarray``.
            n_jobs: Number of processes to read file.
            read_csv_params: Params to read csv file.

        Returns:
            Tuple with read data and new roles mapping.

        """
        spark = get_current_session()

        if isinstance(data, SparkDataFrame):
            out_sdf = data
        elif isinstance(data, str):
            path = cast(str, data)
            if path.endswith(".parquet"):
                out_sdf = spark.read.parquet(path)
                # return pd.read_parquet(data, columns=read_csv_params["usecols"]), None
            elif path.endswith(".csv"):
                csv_params = self._get_read_csv_params()
                out_sdf = spark.read.csv(path, **csv_params)
            else:
                raise ValueError(f"Unsupported data format: {os.path.splitext(path)[1]}")
        else:
            raise ValueError("Input data format is not supported")

        out_sdf = out_sdf.select(*features) if features is not None else out_sdf

        return out_sdf

    def _get_read_csv_params(self) -> Dict:
        return {}

    def _parallel_level(
        self, level: Sequence[SparkMLPipeline], train_valid_iterator: SparkBaseTrainValidIterator
    ) -> Tuple[List[SparkMLPipeline], List[SparkDataset], bool]:
        fit_tasks = [functools.partial(_do_fit, ml_pipe, copy(train_valid_iterator)) for k, ml_pipe in enumerate(level)]

        results = self._computations_manager.compute(fit_tasks)

        ml_pipes, ml_pipes_preds = [list(el) for el in zip(*results)]

        if len(results) < len(level):
            flg_last_level = True
        elif self.timer.time_limit_exceeded() or self.timer.child_out_of_time:
            # if all pipes have been fitted, but generally we don't have time anymore
            flg_last_level = True
        else:
            flg_last_level = False

        return ml_pipes, ml_pipes_preds, flg_last_level


def _do_fit(
    ml_pipe: SparkMLPipeline, iterator: SparkBaseTrainValidIterator
) -> Optional[Tuple[SparkMLPipeline, SparkDataset]]:
    pipe_predictions = cast(SparkDataset, ml_pipe.fit_predict(iterator)).persist(
        level=PersistenceLevel.CHECKPOINT, force=True
    )

    return ml_pipe, pipe_predictions



"""Iterative feature selector."""

import logging

from typing import Iterator
from typing import Optional
from typing import cast

import numpy as np
import pandas as pd

from pandas import Series
from pyspark.sql.pandas.functions import pandas_udf
from pyspark.sql.types import StructField

from sparklightautoml.pipelines.selection.base import SparkImportanceEstimator

from ...computations.base import ComputationsSettings
from ...computations.builder import build_computations_manager
from ...dataset.base import SparkDataset
from ...ml_algo.base import SparkTabularMLAlgo
from ...validation.base import SparkBaseTrainValidIterator


logger = logging.getLogger(__name__)


class SparkNpPermutationImportanceEstimator(SparkImportanceEstimator):
    """Permutation importance based estimator.

    Importance calculate, using random permutation
    of items in single column for each feature.

    """

    def __init__(self, random_state: int = 42, computations_settings: Optional[ComputationsSettings] = None):
        """
        Args:
            random_state: seed for random generation of features permutation.

        """
        super().__init__()
        self.random_state = random_state
        self._computations_manager = build_computations_manager(computations_settings)

    def fit(
        self,
        train_valid: Optional[SparkBaseTrainValidIterator] = None,
        ml_algo: Optional[SparkTabularMLAlgo] = None,
        preds: Optional[SparkDataset] = None,
    ):
        """Find importances for each feature in dataset.

        Args:
            train_valid: Initial dataset iterator.
            ml_algo: Algorithm.
            preds: Predicted target values for validation dataset.

        """
        logger.info(f"Starting importance estimating with {type(self)}")

        assert train_valid is not None, "train_valid cannot be None"

        normal_score = ml_algo.score(preds)
        logger.debug(f"Normal score = {normal_score}")

        valid_data = cast(SparkDataset, train_valid.get_validation_data())

        def build_score_func(it: int, feat: str):
            def func():
                logger.info(f"Start processing ({it},{feat})")
                df = valid_data.data

                field: StructField = df.schema[feat]

                @pandas_udf(returnType=field.dataType)
                def permutate(arrs: Iterator[pd.Series]) -> Iterator[pd.Series]:
                    permutator = np.random.RandomState(seed=self.random_state)
                    # one may get list of arrs and concatenate them to perform permutation
                    # in the whole partition
                    for x in arrs:
                        px = permutator.permutation(x)
                        yield pd.Series(px)

                permutated_df = df.withColumn(feat, permutate(feat))

                ds: SparkDataset = valid_data.empty()
                ds.set_data(permutated_df, valid_data.features, valid_data.roles, name=type(self).__name__)
                logger.debug("Dataframe with shuffled column prepared")

                # Calculate predict and metric
                new_preds = ml_algo.predict(ds)
                shuffled_score = ml_algo.score(new_preds)
                logger.debug(
                    "Shuffled score for col {} = {}, difference with normal = {}".format(
                        feat, shuffled_score, normal_score - shuffled_score
                    )
                )
                return feat, (normal_score - shuffled_score)

            return func

        results = self._computations_manager.compute(
            [build_score_func(it, feat) for it, feat in enumerate(valid_data.features)]
        )

        permutation_importance = {feat: diff_score for feat, diff_score in results}

        self.raw_importances = Series(permutation_importance).sort_values(ascending=False)

        logger.info(f"Finished importance estimating with {type(self)}")



import functools
import importlib
import json
import logging
import os
import uuid
import warnings

from abc import ABC
from abc import abstractmethod
from collections import Counter
from copy import copy
from copy import deepcopy
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from json import JSONDecoder
from json import JSONEncoder
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Set
from typing import Tuple
from typing import Union
from typing import cast

import numpy as np
import pandas as pd

from lightautoml.dataset.base import ColSlice
from lightautoml.dataset.base import IntIdx
from lightautoml.dataset.base import LAMLColumn
from lightautoml.dataset.base import LAMLDataset
from lightautoml.dataset.base import RolesDict
from lightautoml.dataset.base import RowSlice
from lightautoml.dataset.base import array_attr_roles
from lightautoml.dataset.base import valid_array_attributes
from lightautoml.dataset.np_pd_dataset import NpRoles
from lightautoml.dataset.np_pd_dataset import NumpyDataset
from lightautoml.dataset.np_pd_dataset import PandasDataset
from lightautoml.dataset.roles import ColumnRole
from lightautoml.dataset.roles import DateRole
from lightautoml.dataset.roles import DatetimeRole
from lightautoml.dataset.roles import DropRole
from lightautoml.dataset.roles import FoldsRole
from lightautoml.dataset.roles import GroupRole
from lightautoml.dataset.roles import NumericRole
from lightautoml.dataset.roles import PathRole
from lightautoml.dataset.roles import TreatmentRole
from lightautoml.dataset.roles import WeightsRole
from lightautoml.tasks import Task
from pyspark.ml.functions import vector_to_array
from pyspark.sql import Column
from pyspark.sql import functions as sf
from pyspark.sql.session import SparkSession

from sparklightautoml import VALIDATION_COLUMN
from sparklightautoml.dataset.roles import NumericVectorOrArrayRole
from sparklightautoml.utils import SparkDataFrame
from sparklightautoml.utils import create_directory
from sparklightautoml.utils import get_current_session


logger = logging.getLogger(__name__)

Dependency = Union[str, "SparkDataset", "Unpersistable", Callable]
DepIdentifable = Union[str, "SparkDataset"]


class PersistenceLevel(Enum):
    """
    Used for signaling types of persistence points encountered during AutoML process.
    """

    READER = 0
    REGULAR = 1
    CHECKPOINT = 2


class Unpersistable(ABC):
    """
    Interface to provide for external entities to unpersist dataframes and files stored
    by the entity that implements this interface
    """

    def unpersist(self):
        ...


class SparkDatasetMetadataJsonEncoder(JSONEncoder):
    def default(self, o: Any) -> Any:
        if isinstance(o, ColumnRole):
            dct = deepcopy(o.__dict__)
            dct["__type__"] = ".".join([type(o).__module__, type(o).__name__])
            dct["dtype"] = o.dtype.__name__

            if isinstance(o, DatetimeRole):
                dct["date_format"] = dct["format"]
                del dct["format"]
                dct["origin"] = o.origin.timestamp() if isinstance(o.origin, datetime) else o.origin

            return dct

        from sparklightautoml.tasks.base import SparkTask

        if isinstance(o, SparkTask):
            dct = {
                "__type__": type(o).__name__,
                "name": o.name,
                "loss": o.loss_name,
                "metric": o.metric_name,
                # convert to python bool from numpy.bool_
                "greater_is_better": True if o.greater_is_better else False,
            }

            return dct

        raise ValueError(f"Cannot handle object which is not a ColumRole. Object type: {type(o)}")


class SparkDatasetMetadataJsonDecoder(JSONDecoder):
    @staticmethod
    def _column_roles_object_hook(json_object):
        from sparklightautoml.tasks.base import SparkTask

        if json_object.get("__type__", None) == "SparkTask":
            del json_object["__type__"]
            return SparkTask(**json_object)

        if "__type__" in json_object:
            components = json_object["__type__"].split(".")
            module_name = ".".join(components[:-1])
            class_name = components[-1]
            module = importlib.import_module(module_name)
            clazz = getattr(module, class_name)

            del json_object["__type__"]
            if clazz in [DatetimeRole, DateRole]:
                json_object["seasonality"] = tuple(json_object["seasonality"])
                if isinstance(json_object["origin"], float):
                    json_object["origin"] = datetime.fromtimestamp(json_object["origin"])

            json_object["dtype"] = getattr(np, json_object["dtype"])

            if clazz in [GroupRole, DropRole, WeightsRole, FoldsRole, PathRole, TreatmentRole]:
                del json_object["dtype"]

            instance = clazz(**json_object)

            return instance

        return json_object

    def __init__(self, **kwargs):
        kwargs.setdefault("object_hook", self._column_roles_object_hook)
        super().__init__(**kwargs)


class SparkDataset(LAMLDataset, Unpersistable):
    """
    Implements a dataset that uses a ``pyspark.sql.DataFrame`` internally,
    stores some internal state (features, roles, ...) and provide methods to work with dataset.
    """

    @staticmethod
    def _get_rows(data, k: IntIdx) -> Any:
        raise NotImplementedError("This method is not supported")

    @staticmethod
    def _set_col(data: Any, k: int, val: Any):
        raise NotImplementedError("This method is not supported")

    _init_checks = ()
    _data_checks = ()
    _concat_checks = ()
    _dataset_type = "SparkDataset"

    ID_COLUMN = "_id"

    @classmethod
    def load(
        cls,
        path: str,
        file_format: str = "parquet",
        file_format_options: Optional[Dict[str, Any]] = None,
        persistence_manager: Optional["PersistenceManager"] = None,
        partitions_num: Optional[int] = None,
    ) -> "SparkDataset":
        metadata_file_path = os.path.join(path, f"metadata.{file_format}")
        file_path = os.path.join(path, f"data.{file_format}")
        file_format_options = file_format_options or dict()
        spark = get_current_session()

        # reading metadata
        metadata_df = spark.read.format(file_format).options(**file_format_options).load(metadata_file_path)
        metadata = json.loads(
            metadata_df.select("metadata").first().asDict()["metadata"], cls=SparkDatasetMetadataJsonDecoder
        )

        # reading data
        data_df = spark.read.format(file_format).options(**file_format_options).load(file_path)
        name_fixed_cols = (sf.col(c).alias(c.replace("[", "(").replace("]", ")")) for c in data_df.columns)
        data_df = data_df.select(*name_fixed_cols)

        if partitions_num:
            data_df = data_df.repartition(partitions_num)

        return SparkDataset(data=data_df, persistence_manager=persistence_manager, **metadata)

    # TODO: SLAMA - implement filling dependencies
    @classmethod
    def concatenate(
        cls,
        datasets: Sequence["SparkDataset"],
        name: Optional[str] = None,
        extra_dependencies: Optional[List[Dependency]] = None,
    ) -> "SparkDataset":
        """
        Concat multiple datasets by joining their internal ``pyspark.sql.DataFrame``
        using inner join on special hidden '_id' column
        Args:
            datasets: spark datasets to be joined

        Returns:
            a joined dataset, containing features (and columns too) from all datasets
            except containing only one _id column
        """
        assert len(datasets) > 0, "Cannot join an empty list of datasets"

        if len(datasets) == 1:
            return datasets[0]

        if any(not d.bucketized for d in datasets):
            warnings.warn(
                "NOT bucketized datasets are requested to be joined. It may severely affect performance", RuntimeWarning
            )

        # we should join datasets only with unique features
        features = [feat for ds in datasets for feat in ds.features]
        feat_counter = Counter(features)

        assert all(
            count == 1 for el, count in feat_counter.items()
        ), f"Different datasets being joined contain columns with the same names: {feat_counter}"

        roles = {col: role for ds in datasets for col, role in ds.roles.items()}

        except_cols = [c for c in datasets[0].service_columns if c != SparkDataset.ID_COLUMN]
        concatenated_sdf = functools.reduce(
            lambda acc, sdf: acc.join(sdf.drop(*except_cols), on=cls.ID_COLUMN, how="left"), (d.data for d in datasets)
        )

        output = datasets[0].empty()
        output.set_data(
            concatenated_sdf, features, roles, dependencies=[*datasets, *(extra_dependencies or [])], name=name
        )

        return output

    def __init__(
        self,
        data: SparkDataFrame,
        roles: Optional[RolesDict],
        persistence_manager: Optional["PersistenceManager"] = None,
        task: Optional[Task] = None,
        bucketized: bool = False,
        dependencies: Optional[List[Dependency]] = None,
        name: Optional[str] = None,
        target: Optional[str] = None,
        folds: Optional[str] = None,
        **kwargs: Any,
    ):
        self._validate_dataframe(data)

        roles = roles if roles else dict()

        # currently only target is supported
        # adapted from PandasDataset
        for f in roles:
            for k, r in zip(valid_array_attributes, array_attr_roles):
                if roles[f].name == r:
                    roles[f] = DropRole()

        self._data = None
        self._bucketized = bucketized
        self._roles = None
        self._uid = str(uuid.uuid4())
        self._persistence_manager = persistence_manager
        self._dependencies = dependencies
        self._frozen = False
        self._name = name
        self._is_persisted = False
        self._target_column = target
        self._folds_column = folds
        # columns that can be transferred intact across all transformations
        # in the pipeline
        self._service_columns: Set[str] = {self.ID_COLUMN, target, folds, VALIDATION_COLUMN}

        super().__init__(data, list(roles.keys()), roles, task, **kwargs)

    @property
    def uid(self) -> str:
        return self._uid

    @property
    def name(self) -> Optional[str]:
        return self._name

    @property
    def spark_session(self) -> SparkSession:
        return self.data.sql_ctx.sparkSession

    @property
    def data(self) -> SparkDataFrame:
        return self._data

    @data.setter
    def data(self, val: SparkDataFrame) -> None:
        self._data = val

    @property
    def dependencies(self) -> Optional[List[Dependency]]:
        return self._dependencies

    @property
    def features(self) -> List[str]:
        """Get list of features.

        Returns:
            list of features.

        """
        return self._features

    @features.setter
    def features(self, val: List[str]):
        """
        Set features available in this dataset.
        Columns won't be deleted from the dataframe but won't appear throught features property.

        Args:
            val: list of feature names.
        """
        diff = set(val).difference(self.data.columns)
        assert len(diff) == 0, f"Not all roles have features in the dataset. Absent features: {diff}."
        self._features = copy(val)

    @property
    def roles(self) -> RolesDict:
        """Roles dict."""
        return copy(self._roles)

    @roles.setter
    def roles(self, val: NpRoles):
        """Define how to set roles.

        Args:
            val: Roles.

        Note:
            There is different behavior for different type of val parameter:

                - `List` - should be same len as ``data.shape[1]``.
                - `None` - automatic set ``NumericRole(np.float32)``.
                - ``ColumnRole`` - single role for all.
                - ``dict``.

        """
        if type(val) is dict:
            self._roles = dict(((x, val[x]) for x in self.features))
        elif type(val) is list:
            self._roles = dict(zip(self.features, val))
            diff = set(self._roles.keys()).difference(self.data.columns)
            assert len(diff) == 0, f"Not all roles have features in the dataset. Absent features: {diff}."

        elif val:
            role = cast(ColumnRole, val)
            self._roles = dict(((x, role) for x in self.features))
        else:
            raise ValueError()

        diff = set(self._roles.keys()).difference(self.data.columns)
        assert len(diff) == 0, f"Not all roles have features in the dataset. Absent features: {diff}."

    @property
    def bucketized(self) -> bool:
        return self._bucketized

    @property
    def shape(self) -> Tuple[Optional[int], Optional[int]]:
        return -1, len(self.features)
        # with JobGroup("sparkdataset.shape", f"Finding count for dataset (uid={self.uid}, name={self.name})"):
        #     warn_if_not_cached(self.data)
        #     return self.data.count(), len(self.features)

    @property
    def target_column(self) -> Optional[str]:
        return self._target_column

    @property
    def folds_column(self) -> Optional[str]:
        return self._folds_column

    @property
    def service_columns(self) -> List[str]:
        return [sc for sc in self._service_columns if sc in self.data.columns]

    @property
    def num_folds(self) -> int:
        return self.data.select(self.folds_column).distinct().count()

    @property
    def persistence_manager(self) -> "PersistenceManager":
        return self._persistence_manager

    def __repr__(self):
        return f"SparkDataset ({self.data})"

    def __getitem__(self, k: Tuple[RowSlice, ColSlice]) -> Union["LAMLDataset", LAMLColumn]:
        rslice, clice = k

        if isinstance(clice, str):
            clice = [clice]

        assert all(c in self.features for c in clice), (
            f"Not all columns presented in the dataset.\n" f"Presented: {self.features}\n" f"Asked for: {clice}"
        )

        present_svc_cols = [c for c in self.service_columns]
        sdf = cast(SparkDataFrame, self.data.select(*present_svc_cols, *clice))
        roles = {c: self.roles[c] for c in clice}

        output = self.empty()
        output.set_data(sdf, clice, roles, name=self.name)

        return output
        # raise NotImplementedError(f"The method is not supported by {self._dataset_type}")

    def __setitem__(self, k: str, val: Any):
        raise NotImplementedError(f"The method is not supported by {self._dataset_type}")

    def __eq__(self, o: object) -> bool:
        return isinstance(o, SparkDataset) and o.uid == self.uid

    def __hash__(self) -> int:
        return hash(self.uid)

    def _validate_dataframe(self, sdf: SparkDataFrame) -> None:
        assert (
            self.ID_COLUMN in sdf.columns
        ), f"No special unique row id column (the column name: {self.ID_COLUMN}) in the spark dataframe"
        # assert kwargs["target"] in data.columns, \
        #     f"No target column (the column name: {kwargs['target']}) in the spark dataframe"

    def _materialize_to_pandas(
        self,
    ) -> Tuple[pd.DataFrame, Optional[pd.Series], Optional[pd.Series], Dict[str, ColumnRole]]:
        sdf = self.data

        def expand_if_vec_or_arr(col, role) -> Tuple[List[Column], ColumnRole]:
            if not isinstance(role, NumericVectorOrArrayRole):
                return [col], role
            vrole = cast(NumericVectorOrArrayRole, role)

            def to_array(column):
                if vrole.is_vector:
                    return vector_to_array(column)
                return column

            arr = [to_array(sf.col(col))[i].alias(vrole.feature_name_at(i)) for i in range(vrole.size)]

            return arr, NumericRole(dtype=vrole.dtype)

        arr_cols = (expand_if_vec_or_arr(c, self.roles[c]) for c in self.features)
        all_cols_and_roles = {c: role for c_arr, role in arr_cols for c in c_arr}
        all_cols = [scol for scol, _ in all_cols_and_roles.items()]

        if self.target_column is not None:
            all_cols.append(self.target_column)

        if self.folds_column is not None:
            all_cols.append(self.folds_column)

        sdf = sdf.orderBy(SparkDataset.ID_COLUMN).select(*all_cols)
        all_roles = {c: all_cols_and_roles[c] for c in sdf.columns if c not in self.service_columns}

        data = sdf.toPandas()

        df = pd.DataFrame(data=data.to_dict())

        if self.target_column is not None:
            target_series = df[self.target_column]
            df = df.drop(self.target_column, axis=1)
        else:
            target_series = None

        if self.folds_column is not None:
            folds_series = df[self.folds_column]
            df = df.drop(self.folds_column, axis=1)
        else:
            folds_series = None

        return df, target_series, folds_series, all_roles

    def _initialize(self, task: Optional[Task], **kwargs: Any):
        super()._initialize(task, **kwargs)
        self._dependencies = None

    def empty(self) -> "SparkDataset":
        dataset = cast(SparkDataset, super().empty())
        dataset._dependencies = [self]
        dataset._uid = str(uuid.uuid4())
        dataset._frozen = False
        return dataset

    def set_data(
        self,
        data: SparkDataFrame,
        features: List[str],
        roles: NpRoles = None,
        persistence_manager: Optional["PersistenceManager"] = None,
        dependencies: Optional[List[Dependency]] = None,
        uid: Optional[str] = None,
        name: Optional[str] = None,
        frozen: bool = False,
    ):
        """Inplace set data, features, roles for empty dataset.

        Args:
            data: Table with features.
            features: `ignored, always None. just for same interface.
            roles: Dict with roles.
        """
        self._validate_dataframe(data)
        super().set_data(data, features, roles)
        self._persistence_manager = persistence_manager or self._persistence_manager
        self._dependencies = dependencies if dependencies is not None else self._dependencies
        self._uid = uid or self._uid
        self._name = name or self._name
        self._frozen = frozen

    def persist(self, level: Optional[PersistenceLevel] = None, force: bool = False) -> "SparkDataset":
        """
        Materializes current Spark DataFrame and unpersists all its dependencies
        Args:
            level:

        Returns:
            a new SparkDataset that is persisted and materialized
        """
        # if self._is_persisted:
        #     return self

        assert self.persistence_manager, "Cannot persist when persistence_manager is None"

        logger.debug(f"Persisting SparkDataset (uid={self.uid}, name={self.name})")
        level = level if level is not None else PersistenceLevel.REGULAR

        if force:
            ds = self.empty()
            ds.set_data(self.data, self.features, self.roles)
            persisted_dataset = self.persistence_manager.persist(ds, level).to_dataset()
        else:
            persisted_dataset = self.persistence_manager.persist(self, level).to_dataset()

        self._unpersist_dependencies()
        self._is_persisted = True

        return persisted_dataset

    def unpersist(self):
        """
        Unpersists current dataframe if it is persisted and all its dependencies.
        It does nothing if the dataset is frozen
        """
        assert self.persistence_manager, "Cannot unpersist when persistence_manager is None"

        if self.frozen:
            logger.debug(f"Cannot unpersist frozen SparkDataset (uid={self.uid}, name={self.name})")
            return

        logger.debug(f"Unpersisting SparkDataset (uid={self.uid}, name={self.name})")

        self.persistence_manager.unpersist(self.uid)
        self._unpersist_dependencies()

        # if self._is_persisted:
        #     self.persistence_manager.unpersist(self.uid)
        # else:
        #     self._unpersist_dependencies()

    def _unpersist_dependencies(self):
        for dep in self.dependencies or []:
            if isinstance(dep, str):
                self.persistence_manager.unpersist(dep)
            elif isinstance(dep, Unpersistable):
                dep.unpersist()
            else:
                dep()

    @property
    def frozen(self) -> bool:
        return self._frozen

    @frozen.setter
    def frozen(self, val: bool):
        self._frozen = val

    def freeze(self) -> "SparkDataset":
        ds = self.empty()
        ds.set_data(self.data, self.features, self.roles, frozen=True)
        return ds

    def save(
        self,
        path: str,
        save_mode: str = "error",
        file_format: str = "parquet",
        file_format_options: Optional[Dict[str, Any]] = None,
    ):
        metadata_file_path = os.path.join(path, f"metadata.{file_format}")
        file_path = os.path.join(path, f"data.{file_format}")
        file_format_options = file_format_options or dict()

        # prepare metadata of the dataset
        metadata = {
            "name": self.name,
            "roles": self.roles,
            "task": self.task,
            "target": self.target_column,
            "folds": self.folds_column,
        }
        metadata_str = json.dumps(metadata, cls=SparkDatasetMetadataJsonEncoder)
        metadata_df = self.spark_session.createDataFrame([{"metadata": metadata_str}])

        # create directory that will store data and metadata as separate files of dataframes
        create_directory(path, spark=self.spark_session, exists_ok=(save_mode in ["overwrite", "append"]))

        # writing dataframes
        metadata_df.write.format(file_format).mode(save_mode).options(**file_format_options).save(metadata_file_path)
        # fix name of columns: parquet cannot have columns with '(' or ')' in the name
        name_fixed_cols = (sf.col(c).alias(c.replace("(", "[").replace(")", "]")) for c in self.data.columns)
        # write dataframe with fixed column names
        (
            self.data.select(*name_fixed_cols)
            .write.format(file_format)
            .mode(save_mode)
            .options(**file_format_options)
            .save(file_path)
        )

    def to_pandas(self) -> PandasDataset:
        data, target_data, folds_data, roles = self._materialize_to_pandas()

        task = Task(self.task.name) if self.task else None
        kwargs = dict()
        if target_data is not None:
            kwargs["target"] = target_data
        if folds_data is not None:
            kwargs["folds"] = folds_data
        pds = PandasDataset(data=data, roles=roles, task=task, **kwargs)

        return pds

    def to_numpy(self) -> NumpyDataset:
        data, target_data, folds_data, roles = self._materialize_to_pandas()

        try:
            target = self.target
            if isinstance(target, pd.Series):
                target = target.to_numpy()
            elif isinstance(target, SparkDataFrame):
                target = target.toPandas().to_numpy()
        except AttributeError:
            target = None

        try:
            folds = self.folds
            if isinstance(folds, pd.Series):
                folds = folds.to_numpy()
            elif isinstance(folds, SparkDataFrame):
                folds = folds.toPandas().to_numpy()
        except AttributeError:
            folds = None

        return NumpyDataset(
            data=data.to_numpy(), features=list(data.columns), roles=roles, task=self.task, target=target, folds=folds
        )

    @staticmethod
    def _hstack(datasets: Sequence[Any]) -> Any:
        raise NotImplementedError("Unsupported operation for this dataset type")

    @staticmethod
    def _get_cols(data, k: IntIdx) -> Any:
        raise NotImplementedError("Unsupported operation for this dataset type")

    @staticmethod
    def from_dataset(dataset: "LAMLDataset") -> "LAMLDataset":
        assert isinstance(dataset, SparkDataset), "Can only convert from SparkDataset"
        return dataset


@dataclass(frozen=True)
class PersistableDataFrame:
    sdf: SparkDataFrame
    uid: str
    callback: Optional[Callable] = None
    base_dataset: Optional[SparkDataset] = None
    custom_name: Optional[str] = None

    def to_dataset(self) -> SparkDataset:
        assert self.base_dataset is not None
        ds = self.base_dataset.empty()
        ds.set_data(
            self.sdf,
            self.base_dataset.features,
            self.base_dataset.roles,
            dependencies=list(self.base_dataset.dependencies or []),
            uid=self.uid,
            name=self.base_dataset.name,
        )
        return ds

    @property
    def name(self) -> Optional[str]:
        ds_name = self.base_dataset.name if self.base_dataset is not None else None
        return self.custom_name or ds_name


class PersistenceManager(ABC):
    """
    Base interface of an entity responsible for caching and storing intermediate results somewhere.
    """

    @staticmethod
    def to_persistable_dataframe(dataset: SparkDataset) -> PersistableDataFrame:
        # we intentially create new uid to use to distinguish a persisted and unpersisted dataset
        return PersistableDataFrame(dataset.data, uid=dataset.uid, base_dataset=dataset)

    @property
    @abstractmethod
    def uid(self) -> str:
        ...

    @property
    @abstractmethod
    def children(self) -> List["PersistenceManager"]:
        ...

    @property
    @abstractmethod
    def datasets(self) -> List[SparkDataset]:
        ...

    @property
    @abstractmethod
    def all_datasets(self) -> List[SparkDataset]:
        """
        Returns:
            all persisted datasets including persisted with children contexts
        """
        ...

    @abstractmethod
    def persist(
        self, dataset: Union[SparkDataset, PersistableDataFrame], level: PersistenceLevel = PersistenceLevel.REGULAR
    ) -> PersistableDataFrame:
        ...

    @abstractmethod
    def unpersist(self, uid: str):
        ...

    @abstractmethod
    def unpersist_all(self):
        ...

    @abstractmethod
    def unpersist_children(self):
        ...

    @abstractmethod
    def child(self) -> "PersistenceManager":
        ...

    @abstractmethod
    def remove_child(self, child: Union["PersistenceManager", str]):
        ...

    @abstractmethod
    def is_persisted(self, pdf: PersistableDataFrame) -> bool:
        ...



"""
The following methods have been copied from the HandySpark project and modified.
https://github.com/dvgodoy/handyspark
"""
from pyspark.mllib.common import _java2py
from pyspark.mllib.common import _py2java


# noinspection PyProtectedMember
def call2(obj, name, *a):
    serde = obj._sc._jvm.org.apache.spark.mllib.api.python.SerDe
    args = [_py2java(obj._sc, a) for a in a]
    java_res = getattr(obj._java_model, name)(*args)
    java_res = serde.fromTuple2RDD(java_res)
    res = _java2py(obj._sc, java_res)
    return res
